/*
 * Copyright (c) 2005-2007 Extreme! Lab, Indiana University. All rights reserved.
 *
 * This software is open source. See the bottom of this file for the license.
 *
 * $Id: GraphUtil.java,v 1.14 2007/05/22 01:24:26 sshirasu Exp $
 */
package edu.indiana.extreme.xbaya.graph.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import edu.indiana.extreme.xbaya.XBayaRuntimeException;
import edu.indiana.extreme.xbaya.graph.ControlEdge;
import edu.indiana.extreme.xbaya.graph.ControlPort;
import edu.indiana.extreme.xbaya.graph.DataEdge;
import edu.indiana.extreme.xbaya.graph.DataPort;
import edu.indiana.extreme.xbaya.graph.EPRPort;
import edu.indiana.extreme.xbaya.graph.Edge;
import edu.indiana.extreme.xbaya.graph.Graph;
import edu.indiana.extreme.xbaya.graph.GraphException;
import edu.indiana.extreme.xbaya.graph.Node;
import edu.indiana.extreme.xbaya.graph.Port;
import edu.indiana.extreme.xbaya.graph.system.InputNode;
import edu.indiana.extreme.xbaya.graph.system.OutputNode;
import edu.indiana.extreme.xbaya.graph.ws.WSNode;
import edu.indiana.extreme.xbaya.graph.ws.WSPort;
import edu.indiana.extreme.xbaya.gui.ErrorMessages;
import edu.indiana.extreme.xbaya.util.WSConstants;

/**
 * @author Satoshi Shirasuna
 */
public class GraphUtil {

    // private static final MLogger logger = MLogger.getLogger();

    /**
     * Returns the WSNodes included in a specified graph.
     * 
     * @param graph
     *            The specified graph.
     * @return The WSNodes.
     */
    public static Collection<WSNode> getWSNodes(Graph graph) {
        return getNodes(graph, WSNode.class);
    }

    /**
     * Returns a List of InputNodes from a specified graph.
     * 
     * @param graph
     *            the specified graph
     * @return The List of InputNodes.
     */
    public static List<InputNode> getInputNodes(Graph graph) {
        return getNodes(graph, InputNode.class);
    }

    /**
     * Returns a List of OutputNodes from a specified graph.
     * 
     * @param graph
     *            the specified graph
     * @return The List of OutputNodes.
     */
    public static List<OutputNode> getOutputNodes(Graph graph) {
        return getNodes(graph, OutputNode.class);
    }

    /**
     * Returns a List of nodes of specific subclass of Node from a specified
     * graph.
     * 
     * @param <N>
     *            One of the subclass of the Node.
     * @param graph
     *            The specified graph.
     * @param klass
     *            The specified subclass of Node.
     * @return The list of T
     */
    @SuppressWarnings("unchecked")
    public static <N extends Node> List<N> getNodes(Graph graph, Class<N> klass) {
        List<N> nodes = new LinkedList<N>();
        for (Node node : graph.getNodes()) {
            if (klass.isInstance(node)) {
                nodes.add((N) node);
            }
        }
        return nodes;
    }

    /**
     * @param node
     * @return The output nodes.
     */
    public static List<Node> getOutputNodes(Node node) {
        List<Node> outputNodes = new ArrayList<Node>();
        for (Port port : node.getOutputPorts()) {
            Collection<Node> toNodes = port.getToNodes();
            outputNodes.addAll(toNodes);
        }
        return outputNodes;
    }

    /**
     * Returns next nodes connected to a specified node.
     * 
     * @param node
     *            The specified node.
     * @return The next nodes.
     */
    public static List<Node> getNextNodes(Node node) {
        List<Node> nextNodes = getOutputNodes(node);
        for (Port port : node.getControlOutPorts()) {
            Collection<Node> toNodes = port.getToNodes();
            nextNodes.addAll(toNodes);
        }
        return nextNodes;
    }

    /**
     * Sorts the nodes alphabetically by their names.
     * 
     * @param <T>
     * @param nodes
     * @return The list of nodes sorted.
     */
    public static <T extends Node> List<T> sortByName(Collection<T> nodes) {
        List<T> nodeList = new LinkedList<T>(nodes);
        Comparator<Node> nameComparator = new Comparator<Node>() {
            public int compare(Node node1, Node node2) {
                String name1 = node1.getName();
                String name2 = node2.getName();
                return name1.compareToIgnoreCase(name2);
            }
        };
        Collections.sort(nodeList, nameComparator);
        return nodeList;
    }

    /**
     * @param graph
     * @param kind
     * @return The ports of specified kind.
     */
    public static Collection<Port> getPorts(Graph graph, Port.Kind kind) {
        Collection<Port> ports = new ArrayList<Port>();
        for (Port port : graph.getPorts()) {
            if (port.getKind() == kind) {
                ports.add(port);
            }
        }
        return ports;
    }

    /**
     * @param
     * <P>
     * @param graph
     * @param klass
     * @return The ports
     */
    @SuppressWarnings("unchecked")
    public static <P extends Port> List<P> getPorts(Graph graph, Class<P> klass) {
        List<P> ports = new LinkedList<P>();
        for (Port port : graph.getPorts()) {
            if (klass.isInstance(port)) {
                ports.add((P) port);
            }
        }
        return ports;
    }

    private enum Color {
        /**
         * This node hasn't been visited.
         */
        WHITE,
        /**
         * This node has been visited.
         */
        GRAY,
        /**
         * This not is not in cycle.
         */
        BLACK;
    }

    /**
     * @param graph
     * @return true if there is a cycle in the graph; false otherwise.
     */
    public static boolean containsCycle(Graph graph) {
        Map<Node, Color> coloredNodes = new HashMap<Node, Color>();
        for (Node node : graph.getNodes()) {
            coloredNodes.put(node, Color.WHITE);
        }

        for (Node node : graph.getNodes()) {
            if (coloredNodes.get(node) == Color.WHITE) {
                if (visit(node, coloredNodes)) {
                    return true;
                }
            }
        }
        return false;
    }

    private static boolean visit(Node node, Map<Node, Color> coloredNodes) {
        coloredNodes.put(node, Color.GRAY);
        for (Node nextNode : getNextNodes(node)) {
            Color nextNodeColor = coloredNodes.get(nextNode);
            if (nextNodeColor == Color.GRAY) {
                return true;
            } else if (nextNodeColor == Color.WHITE) {
                if (visit(nextNode, coloredNodes)) {
                    return true;
                }
            }
        }
        coloredNodes.put(node, Color.BLACK);
        return false;
    }

    /**
     * @param edge
     * @throws GraphException
     */
    public static void validateConnection(Edge edge) throws GraphException {
        Port fromPort = edge.getFromPort();
        Port toPort = edge.getToPort();
        if (edge instanceof ControlEdge) {
            if (!(fromPort instanceof ControlPort && toPort instanceof ControlPort)) {
                throw new GraphException(ErrorMessages.UNEXPECTED_ERROR);
            }
        } else if (edge instanceof DataEdge) {
            if (fromPort instanceof EPRPort) {
                // TODO
                return;
            }
            if (!(fromPort instanceof DataPort || fromPort instanceof EPRPort)
                    || !(toPort instanceof DataPort)) {
                throw new GraphException(ErrorMessages.UNEXPECTED_ERROR);
            }

            DataPort fromDataPort = (DataPort) fromPort;
            DataPort toDataPort = (DataPort) toPort;

            QName fromType = fromDataPort.getType();
            QName toType = toDataPort.getType();

            if (toDataPort.getEdges().size() > 1) {
                throw new GraphException(
                        ErrorMessages.MORE_THAN_ONE_CONNECTIONS);
            }

            if (!(fromType == null || fromType.equals(WSConstants.XSD_ANY_TYPE)
                    || toType == null
                    || toType.equals(WSConstants.XSD_ANY_TYPE) || fromType
                    .equals(toType))) {
                throw new GraphException(
                        "Cannot connect ports with different types:"
                                + " \nfrom=\t" + fromType + " \nto=\t" + toType
                                + "");
            }
        }
    }

    /**
     * @param graph
     * @throws GraphException
     */
    public static void propagateTypes(Graph graph) throws GraphException {
        List<WSPort> wsPorts = getPorts(graph, WSPort.class);
        for (WSPort wsPort : wsPorts) {
            List<DataEdge> edges = wsPort.getEdges();
            for (DataEdge edge : edges) {
                DataPort fromPort = edge.getFromPort();
                DataPort toPort = edge.getToPort();
                if (fromPort == wsPort) {
                    toPort.copyType(wsPort);
                } else if (toPort == wsPort) {
                    fromPort.copyType(wsPort);
                } else {
                    throw new XBayaRuntimeException();
                }
            }
        }

    }
}

/*
 * Indiana University Extreme! Lab Software License, Version 1.2
 * 
 * Copyright (c) 2005-2007 The Trustees of Indiana University. All rights
 * reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 * 
 * 2) All redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 * 
 * 3) Any documentation included with all redistributions must include the
 * following acknowledgement:
 * 
 * "This product includes software developed by the Indiana University Extreme!
 * Lab. For further information please visit http://www.extreme.indiana.edu/"
 * 
 * Alternatively, this acknowledgment may appear in the software itself, and
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name "Indiana University" or "Indiana University Extreme! Lab" shall
 * not be used to endorse or promote products derived from this software without
 * prior written permission from Indiana University. For written permission,
 * please contact http://www.extreme.indiana.edu/.
 * 
 * 5) Products derived from this software may not use "Indiana University" name
 * nor may "Indiana University" appear in their name, without prior written
 * permission of the Indiana University.
 * 
 * Indiana University provides no reassurances that the source code provided
 * does not infringe the patent or any other intellectual property rights of any
 * other entity. Indiana University disclaims any liability to any recipient for
 * claims brought by any other entity based on infringement of intellectual
 * property rights or otherwise.
 * 
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS.
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS",
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION
 * GENERATED USING SOFTWARE.
 */
